@@ -11,6 +11,8 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 #
+from __future__ import annotations
+
 import datetime
 import json
 from dataclasses import dataclass
@@ -21,20 +23,33 @@ from nats.errors import Error, MsgAlreadyAckdError, NotJSMessageError
 if TYPE_CHECKING:
     from nats import NATS
 
+# Subject without domain:
+# $JS.ACK.<stream>.<consumer>.<delivered>.<sseq>.<cseq>.<tm>.<pending>
+#
+_V1_TOKEN_COUNT = 9
+
+# Subject with domain:
+# $JS.ACK.<domain>.<account hash>.<stream>.<consumer>.<delivered>.<sseq>.
+#   <cseq>.<tm>.<pending>.<a token with a random value>
+#
+_V2_TOKEN_COUNT = 12
+
 
 @dataclass
 class Msg:
     """
     Msg represents a message delivered by NATS.
     """
-    _client: "NATS"
-    subject: str = ''
-    reply: str = ''
-    data: bytes = b''
+
+    _client: NATS
+    subject: str = ""
+    reply: str = ""
+    data: bytes = b""
     headers: Optional[Dict[str, str]] = None
 
-    _metadata: Optional["Metadata"] = None
+    _metadata: Optional[Metadata] = None
     _ackd: bool = False
+    _sid: Optional[int] = None
 
     class Ack:
         Ack = b"+ACK"
@@ -43,8 +58,8 @@ class Msg:
         Term = b"+TERM"
 
         # Reply metadata...
-        Prefix0 = '$JS'
-        Prefix1 = 'ACK'
+        Prefix0 = "$JS"
+        Prefix1 = "ACK"
         Domain = 2
         AccHash = 3
         Stream = 4
@@ -55,32 +70,37 @@ class Msg:
         Timestamp = 9
         NumPending = 10
 
-        # Subject without domain:
-        # $JS.ACK.<stream>.<consumer>.<delivered>.<sseq>.<cseq>.<tm>.<pending>
-        #
-        V1TokenCount = 9
-
-        # Subject with domain:
-        # $JS.ACK.<domain>.<account hash>.<stream>.<consumer>.<delivered>.<sseq>.
-        #   <cseq>.<tm>.<pending>.<a token with a random value>
-        #
-        V2TokenCount = 12
-
     @property
-    def header(self) -> Optional[dict]:
+    def header(self) -> Optional[Dict[str, str]]:
         """
         header returns the headers from a message.
         """
         return self.headers
 
+    @property
+    def is_acked(self) -> bool:
+        """
+        Have we sent a terminal ack message (not in-progress) in response to this original message?
+        """
+        return self._ackd
+
+    @property
+    def sid(self) -> int:
+        """
+        sid returns the subscription ID from a message.
+        """
+        if self._sid is None:
+            raise Error("sid not set")
+        return self._sid
+
     async def respond(self, data: bytes) -> None:
         """
         respond replies to the inbox of the message if there is one.
         """
         if not self.reply:
-            raise Error('no reply subject available')
+            raise Error("no reply subject available")
         if not self._client:
-            raise Error('client not set')
+            raise Error("client not set")
 
         await self._client.publish(self.reply, data, headers=self.headers)
 
@@ -101,7 +121,7 @@ class Msg:
         self._ackd = True
         return resp
 
-    async def nak(self, delay: Optional[Union[int, float]] = None) -> None:
+    async def nak(self, delay: Union[int, float, None] = None) -> None:
         """
         nak negatively acknowledges a message delivered by JetStream triggering a redelivery.
         if `delay` is provided, redelivery is delayed for `delay` seconds
@@ -110,9 +130,9 @@ class Msg:
         payload = Msg.Ack.Nak
         json_args = dict()
         if delay:
-            json_args['delay'] = int(delay * 10**9)  # to seconds to ns
+            json_args["delay"] = int(delay * 10**9)  # from seconds to ns
         if json_args:
-            payload += (b' ' + json.dumps(json_args).encode())
+            payload += b" " + json.dumps(json_args).encode()
         await self._client.publish(self.reply, payload)
         self._ackd = True
 
@@ -121,7 +141,7 @@ class Msg:
         in_progress acknowledges a message delivered by JetStream is still being worked on.
         Unlike other types of acks, an in-progress ack (+WPI) can be done multiple times.
         """
-        if self.reply is None or self.reply == '':
+        if self.reply is None or self.reply == "":
             raise NotJSMessageError
         await self._client.publish(self.reply, Msg.Ack.Progress)
 
@@ -137,70 +157,28 @@ class Msg:
     # TODO(@orsinium): use a cached_property. Available in functools since 3.8,
     # as a package (backports.cached-property), or can be just copy-pasted in the project.
     @property
-    def metadata(self) -> "Metadata":
+    def metadata(self) -> Metadata:
         """
         metadata returns the Metadata of a JetStream message.
         """
-        msg = self
         # Memoize the parsed metadata.
-        metadata = msg._metadata
+        metadata = self._metadata
         if metadata is not None:
             return metadata
-
-        tokens = Msg.Metadata._get_metadata_fields(msg.reply)
-
-        if len(tokens) == Msg.Ack.V1TokenCount:
-            t = datetime.datetime.fromtimestamp(
-                int(tokens[7]) / 1_000_000_000.0
-            )
-            metadata = Msg.Metadata(
-                sequence=Msg.Metadata.SequencePair(
-                    stream=int(tokens[5]),
-                    consumer=int(tokens[6]),
-                ),
-                num_delivered=int(tokens[4]),
-                num_pending=int(tokens[8]),
-                timestamp=t,
-                stream=tokens[2],
-                consumer=tokens[3],
-            )
-        else:
-            t = datetime.datetime.fromtimestamp(
-                int(tokens[Msg.Ack.Timestamp]) / 1_000_000_000.0
-            )
-
-            # Underscore indicate no domain is set. Expose as empty string
-            # to client.
-            domain = tokens[Msg.Ack.Domain]
-            if domain == "_":
-                domain = ""
-
-            metadata = Msg.Metadata(
-                sequence=Msg.Metadata.SequencePair(
-                    stream=int(tokens[Msg.Ack.StreamSeq]),
-                    consumer=int(tokens[Msg.Ack.ConsumerSeq]),
-                ),
-                num_delivered=int(tokens[Msg.Ack.NumDelivered]),
-                num_pending=int(tokens[Msg.Ack.NumPending]),
-                timestamp=t,
-                stream=tokens[Msg.Ack.Stream],
-                consumer=tokens[Msg.Ack.Consumer],
-                domain=domain,
-            )
-
-        msg._metadata = metadata
+        metadata = Msg.Metadata._from_reply(self.reply)
+        self._metadata = metadata
         return metadata
 
     def _get_metadata_fields(self, reply: Optional[str]) -> List[str]:
         return Msg.Metadata._get_metadata_fields(reply)
 
     def _check_reply(self) -> None:
-        if self.reply is None or self.reply == '':
+        if self.reply is None or self.reply == "":
             raise NotJSMessageError
         if self._ackd:
             raise MsgAlreadyAckdError(self)
 
-    @dataclass
+    @dataclass(frozen=True)
     class Metadata:
         """
         Metadata is the metadata from a JetStream message.
@@ -214,19 +192,21 @@ class Msg:
         - consumer is the name of the consumer.
 
         """
-        sequence: Optional["SequencePair"] = None
-        num_pending: Optional[int] = None
-        num_delivered: Optional[int] = None
-        timestamp: Optional[datetime.datetime] = None
-        stream: Optional[str] = None
-        consumer: Optional[str] = None
+
+        sequence: SequencePair
+        num_pending: int
+        num_delivered: int
+        timestamp: datetime.datetime
+        stream: str
+        consumer: str
         domain: Optional[str] = None
 
-        @dataclass
+        @dataclass(frozen=True)
         class SequencePair:
             """
             SequencePair represents a pair of consumer and stream sequence.
             """
+
             consumer: int
             stream: int
 
@@ -234,10 +214,54 @@ class Msg:
         def _get_metadata_fields(cls, reply: Optional[str]) -> List[str]:
             if not reply:
                 raise NotJSMessageError
-            tokens = reply.split('.')
-            if (len(tokens) == Msg.Ack.V1TokenCount or
-                    len(tokens) >= Msg.Ack.V2TokenCount-1) and \
-                    tokens[0] == Msg.Ack.Prefix0 and \
-                    tokens[1] == Msg.Ack.Prefix1:
+            tokens = reply.split(".")
+            if ((len(tokens) == _V1_TOKEN_COUNT
+                 or len(tokens) >= _V2_TOKEN_COUNT - 1)
+                    and tokens[0] == Msg.Ack.Prefix0
+                    and tokens[1] == Msg.Ack.Prefix1):
                 return tokens
             raise NotJSMessageError
+
+        @classmethod
+        def _from_reply(cls, reply: str) -> Msg.Metadata:
+            """Construct the metadata from the reply string"""
+            tokens = cls._get_metadata_fields(reply)
+            if len(tokens) == _V1_TOKEN_COUNT:
+                t = datetime.datetime.fromtimestamp(
+                    int(tokens[7]) / 1_000_000_000.0, datetime.timezone.utc
+                )
+                return cls(
+                    sequence=Msg.Metadata.SequencePair(
+                        stream=int(tokens[5]),
+                        consumer=int(tokens[6]),
+                    ),
+                    num_delivered=int(tokens[4]),
+                    num_pending=int(tokens[8]),
+                    timestamp=t,
+                    stream=tokens[2],
+                    consumer=tokens[3],
+                )
+            else:
+                t = datetime.datetime.fromtimestamp(
+                    int(tokens[Msg.Ack.Timestamp]) / 1_000_000_000.0,
+                    datetime.timezone.utc
+                )
+
+                # Underscore indicate no domain is set. Expose as empty string
+                # to client.
+                domain = tokens[Msg.Ack.Domain]
+                if domain == "_":
+                    domain = ""
+
+                return cls(
+                    sequence=Msg.Metadata.SequencePair(
+                        stream=int(tokens[Msg.Ack.StreamSeq]),
+                        consumer=int(tokens[Msg.Ack.ConsumerSeq]),
+                    ),
+                    num_delivered=int(tokens[Msg.Ack.NumDelivered]),
+                    num_pending=int(tokens[Msg.Ack.NumPending]),
+                    timestamp=t,
+                    stream=tokens[Msg.Ack.Stream],
+                    consumer=tokens[Msg.Ack.Consumer],
+                    domain=domain,
+                )
