@@ -24,7 +24,7 @@ class Error(Exception):
     pass
 
 
-class TimeoutError(asyncio.TimeoutError):
+class TimeoutError(Error, asyncio.TimeoutError):
 
     def __str__(self) -> str:
         return "nats: timeout"
@@ -107,8 +107,10 @@ class SlowConsumerError(Error):
         self.sub = sub
 
     def __str__(self) -> str:
-        return "nats: slow consumer, messages dropped subject: " \
-               f"{self.subject}, sid: {self.sid}, sub: {self.sub}"
+        return (
+            "nats: slow consumer, messages dropped subject: "
+            f"{self.subject}, sid: {self.sid}, sub: {self.sub}"
+        )
 
 
 class BadTimeoutError(Error):
