@@ -12,21 +12,23 @@
 # limitations under the License.
 #
 
-import json
+from __future__ import annotations
+
 import base64
+import json
 from email.parser import BytesParser
-from typing import TYPE_CHECKING, Any, Dict, Optional, List
+from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional
 
 from nats.errors import NoRespondersError
 from nats.js import api
-from nats.js.errors import APIError, ServiceUnavailableError, NotFoundError
+from nats.js.errors import APIError, NotFoundError, ServiceUnavailableError
 
 if TYPE_CHECKING:
     from nats import NATS
 
-NATS_HDR_LINE = bytearray(b'NATS/1.0')
+NATS_HDR_LINE = bytearray(b"NATS/1.0")
 NATS_HDR_LINE_SIZE = len(NATS_HDR_LINE)
-_CRLF_ = b'\r\n'
+_CRLF_ = b"\r\n"
 _CRLF_LEN_ = len(_CRLF_)
 
 
@@ -37,7 +39,7 @@ class JetStreamManager:
 
     def __init__(
         self,
-        conn: "NATS",
+        conn: NATS,
         prefix: str = api.DEFAULT_PREFIX,
         timeout: float = 5,
     ) -> None:
@@ -48,7 +50,7 @@ class JetStreamManager:
 
     async def account_info(self) -> api.AccountInfo:
         resp = await self._api_request(
-            f"{self._prefix}.INFO", b'', timeout=self._timeout
+            f"{self._prefix}.INFO", b"", timeout=self._timeout
         )
         return api.AccountInfo.from_response(resp)
 
@@ -62,19 +64,32 @@ class JetStreamManager:
         info = await self._api_request(
             req_sub, req_data.encode(), timeout=self._timeout
         )
-        return info['streams'][0]
+        if not info["streams"]:
+            raise NotFoundError
+        return info["streams"][0]
 
-    async def stream_info(self, name: str) -> api.StreamInfo:
+    async def stream_info(
+        self,
+        name: str,
+        subjects_filter: Optional[str] = None
+    ) -> api.StreamInfo:
         """
         Get the latest StreamInfo by stream name.
         """
+        req_data = ""
+        if subjects_filter:
+            req_data = json.dumps({"subjects_filter": subjects_filter})
         resp = await self._api_request(
-            f"{self._prefix}.STREAM.INFO.{name}", timeout=self._timeout
+            f"{self._prefix}.STREAM.INFO.{name}",
+            req_data.encode(),
+            timeout=self._timeout,
         )
         return api.StreamInfo.from_response(resp)
 
     async def add_stream(
-        self, config: api.StreamConfig = None, **params
+        self,
+        config: Optional[api.StreamConfig] = None,
+        **params
     ) -> api.StreamInfo:
         """
         add_stream creates a stream.
@@ -82,19 +97,35 @@ class JetStreamManager:
         if config is None:
             config = api.StreamConfig()
         config = config.evolve(**params)
-        if config.name is None:
+
+        stream_name = config.name
+        if stream_name is None:
             raise ValueError("nats: stream name is required")
 
+        # Validate stream name
+        invalid_chars = set(".*>/\\")
+        has_invalid_chars = any(char in stream_name for char in invalid_chars)
+        has_whitespace = any(char.isspace() for char in stream_name)
+        is_not_printable = not stream_name.isprintable()
+
+        if has_invalid_chars or has_whitespace or is_not_printable:
+            raise ValueError(
+                f"nats: stream name ({stream_name}) is invalid. Names cannot contain whitespace, '.', '*', '>', "
+                "path separators (forward or backward slash), or non-printable characters."
+            )
+
         data = json.dumps(config.as_dict())
         resp = await self._api_request(
-            f"{self._prefix}.STREAM.CREATE.{config.name}",
+            f"{self._prefix}.STREAM.CREATE.{stream_name}",
             data.encode(),
             timeout=self._timeout,
         )
         return api.StreamInfo.from_response(resp)
 
     async def update_stream(
-        self, config: api.StreamConfig = None, **params
+        self,
+        config: Optional[api.StreamConfig] = None,
+        **params
     ) -> api.StreamInfo:
         """
         update_stream updates a stream.
@@ -120,25 +151,25 @@ class JetStreamManager:
         resp = await self._api_request(
             f"{self._prefix}.STREAM.DELETE.{name}", timeout=self._timeout
         )
-        return resp['success']
+        return resp["success"]
 
     async def purge_stream(
         self,
         name: str,
         seq: Optional[int] = None,
         subject: Optional[str] = None,
-        keep: Optional[int] = None
+        keep: Optional[int] = None,
     ) -> bool:
         """
         Purge a stream by name.
         """
-        stream_req = {}
+        stream_req: Dict[str, Any] = {}
         if seq:
-            stream_req['seq'] = seq
+            stream_req["seq"] = seq
         if subject:
-            stream_req['filter'] = subject
+            stream_req["filter"] = subject
         if keep:
-            stream_req['keep'] = keep
+            stream_req["keep"] = keep
 
         req = json.dumps(stream_req)
         resp = await self._api_request(
@@ -146,7 +177,7 @@ class JetStreamManager:
             req.encode(),
             timeout=self._timeout
         )
-        return resp['success']
+        return resp["success"]
 
     async def consumer_info(
         self, stream: str, consumer: str, timeout: Optional[float] = None
@@ -156,26 +187,45 @@ class JetStreamManager:
             timeout = self._timeout
         resp = await self._api_request(
             f"{self._prefix}.CONSUMER.INFO.{stream}.{consumer}",
-            b'',
+            b"",
             timeout=timeout
         )
         return api.ConsumerInfo.from_response(resp)
 
-    async def streams_info(self) -> List[api.StreamInfo]:
+    async def streams_info(self, offset=0) -> List[api.StreamInfo]:
         """
-        streams_info retrieves a list of streams.
+        streams_info retrieves a list of streams with an optional offset.
         """
         resp = await self._api_request(
             f"{self._prefix}.STREAM.LIST",
-            b'',
+            json.dumps({
+                "offset": offset
+            }).encode(),
             timeout=self._timeout,
         )
         streams = []
-        for stream in resp['streams']:
+        for stream in resp["streams"]:
             stream_info = api.StreamInfo.from_response(stream)
             streams.append(stream_info)
         return streams
 
+    async def streams_info_iterator(self,
+                                    offset=0) -> Iterable[api.StreamInfo]:
+        """
+        streams_info retrieves a list of streams Iterator.
+        """
+        resp = await self._api_request(
+            f"{self._prefix}.STREAM.LIST",
+            json.dumps({
+                "offset": offset
+            }).encode(),
+            timeout=self._timeout,
+        )
+
+        return api.StreamsListIterator(
+            resp["offset"], resp["total"], resp["streams"]
+        )
+
     async def add_consumer(
         self,
         stream: str,
@@ -193,7 +243,7 @@ class JetStreamManager:
         req_data = json.dumps(req).encode()
 
         resp = None
-        subject = ''
+        subject = ""
         version = self._nc.connected_server_version
         consumer_name_supported = version.major >= 2 and version.minor >= 9
         if consumer_name_supported and config.name:
@@ -215,22 +265,31 @@ class JetStreamManager:
     async def delete_consumer(self, stream: str, consumer: str) -> bool:
         resp = await self._api_request(
             f"{self._prefix}.CONSUMER.DELETE.{stream}.{consumer}",
-            b'',
-            timeout=self._timeout
+            b"",
+            timeout=self._timeout,
         )
-        return resp['success']
+        return resp["success"]
 
-    async def consumers_info(self, stream: str) -> List[api.ConsumerInfo]:
+    async def consumers_info(
+        self,
+        stream: str,
+        offset: Optional[int] = None
+    ) -> List[api.ConsumerInfo]:
         """
-        consumers_info retrieves a list of consumers.
+        consumers_info retrieves a list of consumers. Consumers list limit is 256 for more
+        consider to use offset
+        :param stream: stream to get consumers
+        :param offset: consumers list offset
         """
         resp = await self._api_request(
             f"{self._prefix}.CONSUMER.LIST.{stream}",
-            b'',
+            b"" if offset is None else json.dumps({
+                "offset": offset
+            }).encode(),
             timeout=self._timeout,
         )
         consumers = []
-        for consumer in resp['consumers']:
+        for consumer in resp["consumers"]:
             consumer_info = api.ConsumerInfo.from_response(consumer)
             consumers.append(consumer_info)
         return consumers
@@ -247,25 +306,25 @@ class JetStreamManager:
         get_msg retrieves a message from a stream.
         """
         req_subject = None
-        req = {}
+        req: Dict[str, Any] = {}
         if seq:
-            req['seq'] = seq
+            req["seq"] = seq
         if subject:
-            req['seq'] = None
-            req.pop('seq', None)
-            req['last_by_subj'] = subject
+            req["seq"] = None
+            req.pop("seq", None)
+            req["last_by_subj"] = subject
         if next:
-            req['seq'] = seq
-            req['last_by_subj'] = None
-            req.pop('last_by_subj', None)
-            req['next_by_subj'] = subject
+            req["seq"] = seq
+            req["last_by_subj"] = None
+            req.pop("last_by_subj", None)
+            req["next_by_subj"] = subject
         data = json.dumps(req)
 
         if direct:
             # $JS.API.DIRECT.GET.KV_{stream_name}.$KV.TEST.{key}
-            if subject and not seq:
+            if subject and (seq is None):
                 # last_by_subject type request requires no payload.
-                data = ''
+                data = ""
                 req_subject = f"{self._prefix}.DIRECT.GET.{stream_name}.{subject}"
             else:
                 req_subject = f"{self._prefix}.DIRECT.GET.{stream_name}"
@@ -278,11 +337,11 @@ class JetStreamManager:
 
         # Non Direct form
         req_subject = f"{self._prefix}.STREAM.MSG.GET.{stream_name}"
-        resp = await self._api_request(
+        resp_data = await self._api_request(
             req_subject, data.encode(), timeout=self._timeout
         )
 
-        raw_msg = api.RawStreamMsg.from_response(resp['message'])
+        raw_msg = api.RawStreamMsg.from_response(resp_data["message"])
         if raw_msg.hdrs:
             hdrs = base64.b64decode(raw_msg.hdrs)
             raw_headers = hdrs[NATS_HDR_LINE_SIZE + _CRLF_LEN_:]
@@ -294,10 +353,10 @@ class JetStreamManager:
                     headers[k] = v
             raw_msg.headers = headers
 
-        data = None
+        msg_data: Optional[bytes] = None
         if raw_msg.data:
-            data = base64.b64decode(raw_msg.data)
-        raw_msg.data = data
+            msg_data = base64.b64decode(raw_msg.data)
+        raw_msg.data = msg_data
 
         return raw_msg
 
@@ -305,18 +364,18 @@ class JetStreamManager:
     def _lift_msg_to_raw_msg(self, msg) -> api.RawStreamMsg:
         if not msg.data:
             msg.data = None
-            status = msg.headers.get('Status')
+            status = msg.headers.get("Status")
             if status:
-                if status == '404':
+                if status == "404":
                     raise NotFoundError
                 else:
                     raise APIError.from_msg(msg)
 
         raw_msg = api.RawStreamMsg()
-        subject = msg.headers['Nats-Subject']
+        subject = msg.headers["Nats-Subject"]
         raw_msg.subject = subject
 
-        seq = msg.headers.get('Nats-Sequence')
+        seq = msg.headers.get("Nats-Sequence")
         if seq:
             raw_msg.seq = int(seq)
         raw_msg.data = msg.data
@@ -329,10 +388,10 @@ class JetStreamManager:
         delete_msg retrieves a message from a stream based on the sequence ID.
         """
         req_subject = f"{self._prefix}.STREAM.MSG.DELETE.{stream_name}"
-        req = {'seq': seq}
+        req = {"seq": seq}
         data = json.dumps(req)
         resp = await self._api_request(req_subject, data.encode())
-        return resp['success']
+        return resp["success"]
 
     async def get_last_msg(
         self,
@@ -348,7 +407,7 @@ class JetStreamManager:
     async def _api_request(
         self,
         req_subject: str,
-        req: bytes = b'',
+        req: bytes = b"",
         timeout: float = 5,
     ) -> Dict[str, Any]:
         try:
@@ -358,7 +417,7 @@ class JetStreamManager:
             raise ServiceUnavailableError
 
         # Check for API errors.
-        if 'error' in resp:
-            raise APIError.from_error(resp['error'])
+        if "error" in resp:
+            raise APIError.from_error(resp["error"])
 
         return resp
